##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
from typing import Dict, Tuple, Union, Optional
import os 
import numpy as np
import torch as th

from gbrl.gbrl_wrapper import GBTWrapper, SeparateActorCriticWrapper, SharedActorCriticWrapper
from gbrl.gbt import GBRL
from gbrl.utils import (setup_optimizer, clip_grad_norm, numerical_dtype, 
                    concatenate_arrays, validate_array, constant_like,
                    tensor_to_leaf)


class ActorCritic(GBRL):
    def __init__(self, 
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int, 
                 policy_optimizer: Dict,
                 value_optimizer: Dict= None,
                 shared_tree_struct: bool=True,
                 gbrl_params: Dict=dict(),
                 bias: np.ndarray = None,
                 verbose: int=0,
                 device: str='cpu'):
        
        """ GBRL model for a shared Actor and Critic ensemble.

        Args:
         tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
        output_dim (int): output dimension.
        policy_optimizer Dict: dictionary containing policy optimizer parameters (see GradientBoostingTrees for optimizer details).
        value_optimizer Dict: dictionary containing value optimizer parameters (see GradientBoostingTrees for optimizer details).
        shared_tree_struct (bool, optional): sharing actor and critic. Defaults to True.
        gbrl_params (Dict, optional): GBRL parameters such as:
            control_variates (bool): use control variates (variance reduction technique CPU only).
            split_score_func (str): "cosine" or "l2"
            generator_type- (str): candidate generation method "Quantile" or "Uniform".
            feature_weights - (list[float]): Per-feature multiplication weights used when choosing the best split. Weights should be >= 0
        bias (np.ndarray, optional): manually set a bias. Defaults to None = np.zeros.
        verbose (int, optional): verbosity level. Defaults to 0.
        device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """
        policy_optimizer = setup_optimizer(policy_optimizer, prefix='policy_')
        if value_optimizer is not None:
            value_optimizer = setup_optimizer(value_optimizer, prefix='value_')
        super().__init__(tree_struct,
                         input_dim,
                         output_dim,
                         None,
                         gbrl_params,
                         verbose,
                         device)
        self.policy_optimizer = policy_optimizer
        self.value_optimizer = value_optimizer

        self.shared_tree_struct = True if value_optimizer is None else shared_tree_struct
        self.bias = bias if bias is not None else np.zeros(self.output_dim if shared_tree_struct else self.output_dim - 1, dtype=numerical_dtype)
        # init model
        if self.shared_tree_struct:
            self._model = SharedActorCriticWrapper(self.input_dim, self.output_dim, self.tree_struct, self.policy_optimizer, self.value_optimizer, self.gbrl_params, self.verbose, self.device) 
            self._model.reset()
            self._model.set_bias(self.bias)
        else:
            self._model = SeparateActorCriticWrapper(self.input_dim, self.output_dim, self.tree_struct, self.policy_optimizer, self.value_optimizer, self.gbrl_params, self.verbose, self.device)
            self._model.reset()
            self._model.set_policy_bias(self.bias)
        self.policy_grad = None 
        self.value_grad = None
        
    @classmethod
    def load_model(cls, load_name: str, device: str) -> "ActorCritic":
        """Loads GBRL model from a file

        Args:
            load_name (str): full path to file name

        Returns:
            ActorCritic: loaded ActorCriticModel
        """
        policy_file = load_name + '_policy.gbrl_model'
        value_file = load_name + '_value.gbrl_model'

        instance = cls.__new__(cls)
        if os.path.isfile(policy_file) and os.path.isfile(value_file):
            instance._model = SeparateActorCriticWrapper.load(load_name, device)
            instance.shared_tree_struct = False
            instance.bias = instance._model.policy_model.get_bias() 
        else:
            instance._model = SharedActorCriticWrapper.load(load_name, device)
            instance.shared_tree_struct = True 
            instance.bias = instance._model.get_bias()
        instance.value_optimizer = instance._model.value_optimizer
        instance.policy_optimizer = instance._model.policy_optimizer
        instance.input_dim = instance._model.input_dim
        instance.output_dim = instance._model.output_dim
        instance.verbose = instance._model.verbose
        instance.tree_struct = instance._model.tree_struct
        instance.gbrl_params = instance._model.gbrl_params
        instance.device = instance._model.get_device()
        if isinstance(instance.device, tuple):
            instance.device = instance.device[0]
        return instance
    
    def get_num_trees(self) -> Union[int, Tuple[int, int]]:
        """ Returns number of trees in the ensemble.
        If separate actor and critic return number of trees per ensemble.
        Returns:
            Union[int, Tuple[int, int]]
        """
        return self._model.get_num_trees()
     
    def predict_values(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> Union[np.ndarray, th.Tensor]:
        """Predict only values. If `requires_grad=True` then stores 
           differentiable parameters in self.params 
           Return type/device is identical to the input type/device.

        Args:
            observations (Union[np.ndarray, th.Tensor])
            requires_grad (bool, optional). Defaults to True. Ignored if input is a numpy array.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Union[np.ndarray, th.Tensor]: values
        """
        values = self._model.predict_critic(observations, requires_grad, start_idx, stop_idx, tensor)
        if requires_grad:
            self.value_grad = None
            self.params = values
        return values

    def __call__(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> Tuple[Union[np.ndarray, th.Tensor], Union[np.ndarray, th.Tensor]]:
        """ Predicts  and returns actor and critic outputs as tensors. If `requires_grad=True` then stores 
           differentiable parameters in self.params 
           Return type/device is identical to the input type/device.
        Args:
            observations (Union[np.ndarray, th.Tensor])
            requires_grad (bool, optional). Defaults to True. Ignored if input is a numpy array.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Tuple[Union[np.ndarray, th.Tensor], Union[np.ndarray, th.Tensor]]: actor and critic output
        """
        params = self._model.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        if requires_grad:
            self.policy_grad = None
            self.value_grad = None
            self.params = params
            self.input = observations
        return params
    
    def step(self, observations: Optional[Union[np.ndarray, th.Tensor]] = None, policy_grad: Optional[Union[np.ndarray, th.Tensor]] = None, value_grad: Optional[Union[np.ndarray, th.Tensor]] = None, policy_grad_clip: Optional[float] = None, value_grad_clip : Optional[float] = None) -> None:
        """Performs a boosting step for both actor and critic

        Args:
            observations (Union[np.ndarray, th.Tensor]):
            policy_grad_clip (float, optional): . Defaults to None.
            value_grad_clip (float, optional):. Defaults to None.
            policy_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.
            value_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.
        """
        if observations is None:
            assert self.input is not None, "Cannot update trees without input. Make sure model is called with requires_grad=True"
            observations = self.input
        n_samples = len(observations)

        policy_grad = policy_grad if policy_grad is not None else self.params[0].grad.detach() * n_samples
        value_grad = value_grad if value_grad is not None else self.params[1].grad.detach() * n_samples

        policy_grad = clip_grad_norm(policy_grad, policy_grad_clip)
        value_grad = clip_grad_norm(value_grad, value_grad_clip)

        validate_array(policy_grad)
        validate_array(value_grad)

        self._model.step(observations, policy_grad, value_grad)
        self.policy_grad = policy_grad
        self.value_grad = value_grad
        self.input = None
    
    def actor_step(self, observations: Optional[Union[np.ndarray, th.Tensor]] = None, policy_grad: Optional[Union[np.ndarray, th.Tensor]] = None, policy_grad_clip: Optional[float] = None) -> None:
        """Performs a single boosting step for the actor (should only be used if actor and critic use separate models)

        Args:
            observations (Union[np.ndarray, th.Tensor]):
            policy_grad_clip (float, optional): Defaults to None.
            policy_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.

        Returns:
            np.ndarray: policy gradient
        """
        assert not self.shared_tree_struct, "Cannot separate boosting steps for actor and critic when using separate tree architectures!"
        if observations is None:
            assert self.input is not None, "Cannot update trees without input. Make sure model is called with requires_grad=True"
            observations = self.input
        n_samples = len(observations)
        policy_grad = policy_grad if policy_grad is not None else self.params[0].grad.detach() * n_samples
        policy_grad = clip_grad_norm(policy_grad, policy_grad_clip)
        validate_array(policy_grad)

        self._model.step_policy(observations, policy_grad)
        self.policy_grad = policy_grad
    
    def critic_step(self, observations: Optional[Union[np.ndarray, th.Tensor]] = None, value_grad: Optional[Union[np.ndarray, th.Tensor]] = None, value_grad_clip: Optional[float] = None) -> None:
        """Performs a single boosting step for the critic (should only be used if actor and critic use separate models)

        Args:
            observations (Union[np.ndarray, th.Tensor]):
            value_grad_clip (float, optional): Defaults to None.
            value_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.

        Returns:
            np.ndarray: value gradient
        """
        assert not self.shared_tree_struct, "Cannot separate boosting steps for actor and critic when using separate tree architectures!"
        if observations is None:
            assert self.input is not None, "Cannot update trees without input. Make sure model is called with requires_grad=True"
            observations = self.input
        n_samples = len(observations)
        
        value_grad = value_grad if value_grad is not None else self.params[1].grad.detach() * n_samples
        value_grad = clip_grad_norm(value_grad, value_grad_clip)

        validate_array(value_grad)
        self._model.step_critic(observations, value_grad)
        self.value_grad = value_grad

    def get_params(self) -> Tuple[np.ndarray, np.ndarray]:
        """Returns predicted actor and critic parameters and their respective gradients

        Returns:
            Tuple[np.ndarray, np.ndarray]
        """
        assert self.params is not None, "must run a forward pass first"
        if isinstance(self.params, tuple):
            return (self.params[0].detach().cpu().numpy(),self.params[1].detach().cpu().numpy()) , (self.policy_grad, self.value_grad)
        return self.params, (self.policy_grad, self.value_grad)
    
    def copy(self) -> "ActorCritic":
        """Copy class instance 

        Returns:
            ActorCritic: copy of current instance
        """
        return self.__copy__()

    def __copy__(self) -> "ActorCritic":
        value_optimizer = None if self.value_optimizer is None else self.value_optimizer.copy()
        copy_ = ActorCritic(self.tree_struct.copy(), self.input_dim, self.output_dim, self.policy_optimizer.copy(), value_optimizer, self.shared_tree_struct, self.gbrl_params, self.bias, self.verbose, self.device)
        if self._model is not None:
            copy_._model = self._model.copy()
        return copy_
    
class ParametricActor(GBRL):
    def __init__(self, 
                 tree_struct: Dict,
                 input_dim: int, 
                 output_dim: int, 
                 policy_optimizer: Dict,
                 gbrl_params: Dict=dict(),
                 bias: np.ndarray = None,
                 verbose: int=0,
                 device: str='cpu'):
        """ GBRL model for a ParametricActor ensemble. ParametricActor outputs a single parameter per action dimension.
            Therefore it can be determinstic or stochastic (e.g. Discrete action space).

        Args:
         tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
        output_dim (int): output dimension.
        policy_optimizer Dict: dictionary containing policy optimizer parameters (see GradientBoostingTrees for optimizer details).
        gbrl_params (Dict, optional): GBRL parameters such as:
            control_variates (bool): use control variates (variance reduction technique CPU only).
            split_score_func (str): "cosine" or "l2"
            generator_type- (str): candidate generation method "Quantile" or "Uniform".
            feature_weights - (list[float]): Per-feature multiplication weights used when choosing the best split. Weights should be >= 0
        bias (np.ndarray, optional): manually set a bias. Defaults to None = np.zeros.
        verbose (int, optional): verbosity level. Defaults to 0.
        device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """
        policy_optimizer = setup_optimizer(policy_optimizer, prefix='policy_')
        super().__init__(tree_struct,
                         input_dim,
                         output_dim,
                         None,
                         gbrl_params,
                         verbose,
                         device)
        self.policy_optimizer = policy_optimizer
        self.bias = bias if bias is not None else np.zeros(self.output_dim, dtype=numerical_dtype)
        # init model
        self._model = GBTWrapper(self.output_dim, self.tree_struct, self.policy_optimizer, self.gbrl_params, self.verbose, self.device) 
        self._model.reset()
        self._model.set_bias(self.bias)

    def step(self, observations: Optional[Union[np.ndarray, th.Tensor]] = None, policy_grad: Optional[Union[np.ndarray, th.Tensor]] = None, policy_grad_clip: Optional[float] = None,) -> None:
        """Performs a single boosting iteration.

        Args:
            observations (Union[np.ndarray, th.Tensor]):
            policy_grad_clip (float, optional): . Defaults to None.
            policy_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.
        """
        if observations is None:
            assert self.input is not None, "Cannot update trees without input. Make sure model is called with requires_grad=True"
            observations = self.input
        n_samples = len(observations)
            
        policy_grad = policy_grad if policy_grad is not None else self.params.grad.detach() * n_samples
        policy_grad = clip_grad_norm(policy_grad, policy_grad_clip)
        validate_array(policy_grad)

        self._model.step(observations, policy_grad)
        self.grad = policy_grad
        self.input = None

    @classmethod
    def load_model(cls, load_name: str, device: str) -> "ParametricActor":
        """Loads GBRL model from a file

        Args:
            load_name (str): full path to file name

        Returns:
            ParametricActor: loaded ActorCriticModel
        """
        instance = cls.__new__(cls)
        instance._model = GBTWrapper.load(load_name, device)
        instance.bias = instance._model.get_bias()
        instance.policy_optimizer = instance._model.optimizer
        instance.input_dim = instance._model.input_dim
        instance.output_dim = instance._model.output_dim
        instance.verbose = instance._model.verbose
        instance.tree_struct = instance._model.tree_struct
        instance.gbrl_params = instance._model.gbrl_params
        instance.device = instance._model.get_device()
        return instance
    
    def get_num_trees(self) -> int:
        """ Returns number of trees in the ensemble.
        Returns:
            int
        """
        return self._model.get_num_trees()
     
    def __call__(self, observations: Union[np.ndarray, th.Tensor], requires_grad : bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> Union[np.ndarray, th.Tensor]:
        """ Returns actor output as Tensor. If `requires_grad=True` then stores 
           differentiable parameters in self.params 
           Return type/device is identical to the input type/device.
           Requires_grad is ignored if input is a numpy array.
        Args:
            observations (Union[np.ndarray, th.Tensor])
            requires_grad bool: Defaults to None.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Union[np.ndarray, th.Tensor]: GBRL outputs - a single parameter per action dimension.
        """
        params = self._model.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        if requires_grad:
            self.grads = None
            self.params = params
            self.input = observations
        return params
    
    def __copy__(self) -> "ParametricActor":
        copy_ = ParametricActor(self.tree_struct.copy(), self.input_dim, self.output_dim, self.policy_optimizer.copy(),  self.gbrl_params, self.bias, self.verbose, self.device)
        if self._model is not None:
            copy_._model = self._model.copy()
        return copy_
    
class GaussianActor(GBRL):
    def __init__(self, 
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int, 
                 mu_optimizer: Dict,
                 std_optimizer: Dict = None,
                 log_std_init: float = -2,
                 gbrl_params: Dict = dict(),
                 bias: np.ndarray = None,
                 verbose: int=0,
                 device: str='cpu'):
        """ GBRL model for a Actor ensemble used in algorithms such as: SAC.
        Model outputs mu and log_std of a Gaussian distribution. 
        Actor optimizer can be shared for both parameters or separate. 
        
        Args:
         tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
        output_dim (int): output dimension.
        mu_optimizer Dict: dictionary containing Gaussian mean optimizer parameters. (see GradientBoostingTrees for optimizer details)
        std_optimizer Dict: dictionary containing Gaussian sigma optimizer parameters. (see GradientBoostingTrees for optimizer details)
        log_std_init (float): initial value of log_std
        gbrl_params (Dict, optional): GBRL parameters such as:
            control_variates (bool): use control variates (variance reduction technique CPU only).
            split_score_func (str): "cosine" or "l2"
            generator_type- (str): candidate generation method "Quantile" or "Uniform".
            feature_weights - (list[float]): Per-feature multiplication weights used when choosing the best split. Weights should be >= 0
        bias (np.ndarray, optional): manually set a bias. Defaults to None = np.zeros.
        verbose (int, optional): verbosity level. Defaults to 0.
        device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """
        mu_optimizer = setup_optimizer(mu_optimizer, prefix='mu_')

        self.bias = bias if bias is not None else np.zeros(output_dim, dtype=numerical_dtype)

        policy_dim = output_dim
        if std_optimizer is not None:
            std_optimizer = setup_optimizer(std_optimizer, prefix='std_')
            policy_dim = output_dim // 2
            self.bias[policy_dim:] = log_std_init*np.ones(policy_dim, dtype=numerical_dtype)
        
        self.mu_optimizer = mu_optimizer
        self.std_optimizer = std_optimizer
        self.log_std_init = log_std_init

        super().__init__(tree_struct,
                         input_dim,
                         output_dim,
                         None, 
                         gbrl_params,
                         verbose,
                         device)

        
        self.policy_dim = policy_dim
        # init model
        self._model = GBTWrapper(self.input_dim, self.output_dim, self.tree_struct, [mu_optimizer, std_optimizer], self.gbrl_params, self.verbose, self.device)
        self._model.reset()
        self._model.set_bias(self.bias)
        
    def step(self, observations: Optional[Union[np.ndarray, th.Tensor]] = None, mu_grad: Optional[Union[np.ndarray, th.Tensor]] = None, log_std_grad: Optional[Union[np.ndarray, th.Tensor]] = None, mu_grad_clip: Optional[float] = None, log_std_grad_clip: Optional[float] = None) -> None:
        """Performs a single boosting iteration.

        Args:
            observations (Union[np.ndarray, th.Tensor])
            mu_grad_clip (float, optional). Defaults to None.
            log_std_grad_clip (float, optional). Defaults to None.
            mu_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.
            log_std_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.
        """
        if observations is None:
            assert self.input is not None, "Cannot update trees without input. Make sure model is called with requires_grad=True"
            observations = self.input
        n_samples = len(observations)
        mu_grad = mu_grad if mu_grad is not None else self.params[0].grad.detach() * n_samples
        mu_grad = clip_grad_norm(mu_grad, mu_grad_clip)
  
        if self.std_optimizer is not None:
            log_std_grad = log_std_grad if log_std_grad is not None else self.params[1].grad.detach() * n_samples
            log_std_grad = clip_grad_norm(log_std_grad, log_std_grad_clip)
            theta_grad = concatenate_arrays(mu_grad, log_std_grad)
        else:
            theta_grad = mu_grad
        
        validate_array(theta_grad)

        self._model.step(observations, theta_grad)
        self.grad = mu_grad
        if self.std_optimizer is not None:
            self.grad = (mu_grad, log_std_grad)
        self.input = None
    
    def get_num_trees(self) -> int:
        """Returns the number of trees in the ensemble

        Returns:
            int: number of trees
        """
        return self._model.get_num_trees()

    def __call__(self, observations: Union[np.ndarray, th.Tensor], requires_grad : bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> Union[np.ndarray, th.Tensor]:
        """Returns actor's outputs as tensor. If `requires_grad=True` then stores 
           differentiable parameters in self.params. Return type/device is identical to the input type/device.
           Requires_grad is ignored if input is a numpy array.
        Args:
            observations (Union[np.ndarray, th.Tensor])
            requires_grad bool: Defaults to None.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Union[np.ndarray, th.Tensor]: Gaussian parameters
        """
        theta = self._model.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        mean_actions = theta if self.std_optimizer is None else theta[:, :self.policy_dim]
        log_std = constant_like(theta, self.log_std_init) if self.std_optimizer is None else theta[:, self.policy_dim:]
        mean_actions = tensor_to_leaf(mean_actions, requires_grad)
        log_std = tensor_to_leaf(log_std, requires_grad = False if self.std_optimizer is None else requires_grad)
        if requires_grad:
            self.grad = None
            self.params = mean_actions, log_std
            self.input = observations
        return mean_actions, log_std
    
    def __copy__(self) -> "GaussianActor":
        std_optimizer = None if self.std_optimizer is None else self.std_optimizer.copy()
        copy_ = GaussianActor(self.tree_struct.copy(), self.input_dim, self.output_dim, self.mu_optimizer.copy(), std_optimizer, self.gbrl_params, self.bias, self.verbose, self.device)
        if self._model is not None:
            copy_._model = self._model.copy()
        return copy_
    

class ContinuousCritic(GBRL):
    def __init__(self, 
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int, 
                 weights_optimizer: Dict,
                 bias_optimizer: Dict =  None,
                 gbrl_params: Dict=dict(),
                 target_update_interval: int = 100,
                 bias: np.ndarray = None,
                 verbose: int=0,
                 device: str='cpu'):         
        """ GBRL model for a Continuous Critic ensemble.
            Usage example Q-function in SAC.
            Model is designed to output parameters of 3 types of Q-functions:
                - linear Q(theta(s), a) = <w_theta, a> + b_theta, (<> denotes a dot product).
                - quadratic Q(theta(s), a) = -(<w_theta, a> - b_theta)**2 + c_theta  
                - tanh Q(theta(s), a) = b_theta*tanh(<w_theta, a>)
            This allows to pass derivatives w.r.t to action a while the Q parameters are a function of a GBT model theta. 
            The target model is approximated as the ensemble without the last <target_update_interval> trees.
        
        Args:
         tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
        output_dim (int): output dimension.
        weights_optimizer Dict: dictionary containing policy optimizer parameters. (see GradientBoostingTrees for optimizer details)
        bias_optimizer Dict: dictionary containing policy optimizer parameters. (see GradientBoostingTrees for optimizer details)
        gbrl_params (Dict, optional): GBRL parameters such as:
            control_variates (bool): use control variates (variance reduction technique CPU only).
            split_score_func (str): "cosine" or "l2"
            generator_type- (str): candidate generation method "Quantile" or "Uniform".
            feature_weights - (list[float]): Per-feature multiplication weights used when choosing the best split. Weights should be >= 0
        target_update_interval (int): target update interval
        bias (np.ndarray, optional): manually set a bias. Defaults to None = np.zeros.
        verbose (int, optional): verbosity level. Defaults to 0.
        device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """

        weights_optimizer = setup_optimizer(weights_optimizer, prefix='weights_')
        bias_optimizer = setup_optimizer(bias_optimizer, prefix='bias_')

        super().__init__(tree_struct,
                         input_dim,
                         output_dim,
                         None, 
                         gbrl_params,
                         verbose,
                         device)
        self.weights_optimizer = weights_optimizer
        self.bias_optimizer = bias_optimizer
        self.target_model = None
        self.bias = bias if bias is not None else np.zeros(self.output_dim, dtype=numerical_dtype)
        self.target_update_interval = target_update_interval
        # init model
        self._model = GBTWrapper(self.input_dim, self.output_dim, self.tree_struct, [weights_optimizer, bias_optimizer], self.gbrl_params, self.verbose, self.device)
        self._model.reset()
        self._model.set_bias(self.bias)
        
    def step(self, observations: Optional[Union[np.ndarray, th.Tensor]] = None, weight_grad: Optional[Union[np.ndarray, th.Tensor]] = None, bias_grad: Optional[Union[np.ndarray, th.Tensor]] = None, q_grad_clip: Optional[float] = None) -> None:
        """Performs a single boosting step

        Args:
            observations (Union[np.ndarray, th.Tensor]):
            q_grad_clip (float, optional):. Defaults to None.
        weight_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.
            bias_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.           
        """
        if observations is None:
            assert self.input is not None, "Cannot update trees without input. Make sure model is called with requires_grad=True"
            observations = self.input
        n_samples = len(observations)
        weight_grad = weight_grad if weight_grad is not None else self.params[0].grad.detach() * n_samples
        bias_grad = bias_grad if bias_grad is not None else self.params[1].grad.detach() * n_samples

        weight_grad = clip_grad_norm(weight_grad, q_grad_clip)
        bias_grad = clip_grad_norm(bias_grad, q_grad_clip)

        validate_array(weight_grad)
        validate_array(bias_grad)
        theta_grad = concatenate_arrays(weight_grad, bias_grad)

        self._model.step(observations, theta_grad)
        self.grad = (weight_grad, bias_grad)
        self.input = None

    def predict_target(self, observations: Union[np.ndarray, th.Tensor], tensor: bool = True) -> Tuple[Union[np.ndarray, th.Tensor], Union[np.ndarray, th.Tensor]]:
        """Predict the parameters of a Target Continuous Critic as Tensors.
        Prediction is made by summing the outputs the trees from Continuous Critic model up to `n_trees - target_update_interval`.

        Args:
            observations (Union[np.ndarray, th.Tensor]):
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Tuple[th.Tensor, th.Tensor]: weights and bias parameters to the type of Q-functions
        """
        n_trees = self._model.get_num_trees()
        theta = self._model.predict(observations, requires_grad=False, stop_idx=max(n_trees - self.target_update_interval, 1), tensor=tensor)
        weights, bias = theta[:, self.weights_optimizer['start_idx']:self.weights_optimizer['stop_idx']], theta[:, self.bias_optimizer['start_idx']:self.bias_optimizer['stop_idx']]
        return weights, bias

    def get_num_trees(self) -> int:
        """Get number of trees in model.

        Returns:
            int: return number of trees
        """
        return self._model.get_num_trees()

    def __call__(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool =  True, target: bool = False, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> Tuple[Union[np.ndarray, th.Tensor], Union[np.ndarray, th.Tensor]]:
        """Predict the parameters of a Continuous Critic as Tensors. if `requires_grad=True` then stores 
           differentiable parameters in self.params. 
           Return type/device is identical to the input type/device.

        Args:
            observations (Union[np.ndarray, th.Tensor])
            requires_grad (bool, optional). Defaults to True.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Tuple[Union[np.ndarray, th.Tensor], Union[np.ndarray, th.Tensor]]: weights and bias parameters to the type of Q-functions
        """
        if target: 
            return self.predict_target(observations, tensor)

        theta = self._model.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        weights, bias = theta[:, self.weights_optimizer['start_idx']:self.weights_optimizer['stop_idx']], theta[:, self.bias_optimizer['start_idx']:self.bias_optimizer['stop_idx']]
        if requires_grad:
            self.grad = None
            self.params = weights, bias
            self.input = observations
        return weights, bias
    
    def __copy__(self) -> "ContinuousCritic":
        copy_ = ContinuousCritic(self.tree_struct.copy(), self.input_dim, self.output_dim, self.weights_optimizer.copy(), self.bias_optimizer.copy() if isinstance(self.critic_optimizer, dict) else {'weights_optimizer': self.critic_optimizer[0], 'bias_optimizer': self.critic_optimizer[1]}, self.gbrl_params, self.target_update_interval, self.bias, self.verbose, self.device)
        if self._model is not None:
            copy_._model = self._model.copy()
        return copy_
    

class DiscreteCritic(GBRL):
    def __init__(self, 
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int, 
                 critic_optimizer: Dict,
                 gbrl_params: Dict=dict(),
                 target_update_interval: int = 100,
                 bias: np.ndarray = None,
                 verbose: int=0,
                 device: str='cpu'):  
        """ GBRL model for a Discrete Critic ensemble.
            Usage example: Q-function in GBRL.
            The target model is approximated as the ensemble without the last <target_update_interval> trees.
        Args:
         tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
        output_dim (int): output dimension.
        critic_optimizer Dict: dictionary containing policy optimizer parameters. (see GradientBoostingTrees for optimizer details).
        gbrl_params (Dict, optional): GBRL parameters such as:
            control_variates (bool): use control variates (variance reduction technique CPU only).
            split_score_func (str): "cosine" or "l2"
            generator_type- (str): candidate generation method "Quantile" or "Uniform".
            feature_weights - (list[float]): Per-feature multiplication weights used when choosing the best split. Weights should be >= 0
        verbose (int, optional): verbosity level. Defaults to 0.
        device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """
        critic_optimizer = setup_optimizer(critic_optimizer, prefix='critic_')
        super().__init__(tree_struct,
                         input_dim,
                         output_dim,
                         None, 
                         gbrl_params,
                         verbose,
                         device)
        self.critic_optimizer = critic_optimizer
        self.target_update_interval = target_update_interval
        self.bias = bias if bias is not None else np.zeros(self.output_dim, dtype=numerical_dtype)
        # init model
        self._model = GBTWrapper(self.input_dim, self.output_dim, self.tree_struct, self.critic_optimizer, self.gbrl_params, self.verbose, self.device)
        self._model.reset()
        self._model.set_bias(self.bias)

    def step(self, observations: Optional[Union[np.ndarray, th.Tensor]] = None, q_grad: Optional[Union[np.ndarray, th.Tensor]] = None, max_q_grad_norm: Optional[np.ndarray] = None) -> None:
        """Performs a single boosting iterations.

        Args:
            observations (Union[np.ndarray, th.Tensor]):
            max_q_grad_norm (np.ndarray, optional). Defaults to None.
            q_grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.
        """
        if observations is None:
            assert self.input is not None, "Cannot update trees without input. Make sure model is called with requires_grad=True"
            observations = self.input
        if q_grad is None:
            n_samples = len(observations)
            q_grad = self.params.grad.detach().cpu().numpy() * n_samples
        q_grad = clip_grad_norm(q_grad, max_q_grad_norm)
      
        self._model.step(observations, q_grad)
        self.grad = q_grad
        self.input = None

    def __call__(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: Optional[int] = None, tensor: bool = True) -> Union[np.ndarray, th.Tensor]:
        """Predict and return Critic's outputs as Tensors. if `requires_grad=True` then stores 
           differentiable parameters in self.params. 
           Return type/device is identical to the input type/device.

        Args:
            observations (Union[np.ndarray, th.Tensor])
            requires_grad (bool, optional). Defaults to True.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Union[np.ndarray, th.Tensor]: Critic's outputs.
        """
        q_values =self._model.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        if requires_grad:
            self.grad = None
            self.params = q_values
            self.input = observations
        return q_values

    def predict_target(self, observations: Union[np.ndarray, th.Tensor], tensor: bool = True) -> th.Tensor:
        """Predict and return Target Critic's outputs as Tensors.
           Prediction is made by summing the outputs the trees from Continuous Critic model up to `n_trees - target_update_interval`.
        
        Args:
            observations (Union[np.ndarray, th.Tensor])
        
        Returns:
            th.Tensor: Target Critic's outputs.
        """
        n_trees = self._model.get_num_trees()
        return self._model.predict(observations, requires_grad=False, stop_idx=max(n_trees - self.target_update_interval, 1), tensor=tensor)

    def __copy__(self) -> "DiscreteCritic":
        copy_ = DiscreteCritic(self.tree_struct.copy(), self.input_dim, self.output_dim, self.critic_optimizer.copy(), self.gbrl_params, self.target_update_interval, self.bias, self.verbose, self.device)
        if self._model is not None:
            copy_._model = self._model.copy()
        return copy_

if __name__ == '__main__':
    pass
        







##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
from typing import Dict, List, Union, Tuple, Optional, Any

import numpy as np
import torch as th


from gbrl.gbrl_wrapper import GBTWrapper
from gbrl.utils import setup_optimizer, clip_grad_norm, validate_array


class GBRL:
    def __init__(self, 
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int,
                 optimizer: Union[Dict, List[Dict]],
                 gbrl_params: Dict = dict(),
                 verbose: int=0,
                 device: str = 'cpu'):
        """General class for gradient boosting trees

        Args:
            tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
            output_dim (int): output dimension.
            optimizer (Union[Dict, List[Dict]]): dictionary containing optimizer parameters or a list of dictionaries containing optimizer parameters.
                SGD optimizer must contain as:
                    'algo': 'SGD'
                    'lr' Union[float, str]: learning rate value. Constant learning rate is used by default
                Adam optimizer must contain (CPU only):
                    'algo' (str): 'Adam'
                    'lr' Union[float, str]: learning rate value. Constant learning rate is used by default
                    'beta_1' (float): 0.99
                    'beta_2' (float): 0.999
                    'eps' (float): 1.0e-8
                All optimizers must contain:
                    'start_idx' 
                    'stop_idx'
                Setting scheduler type: 
                Available schedulers are Constant and Linear. Constant is default, Linear is CPU only.
                To specify a linear scheduler, 3 additional arguments must be added to an optimizer dict.
                    'init_lr' (str): "lin_<value>"
                    'stop_lr' (float): minimum lr value 
                    'T' (int): number of total expected boosting trees, 
                               used to calculate the linear scheduling internally.
                               Can be manually calculated per algorithm according 
                               to the total number of training steps.
             gbrl_params (Dict, optional): GBRL parameters such as:
                control_variates (bool): use control variates (variance reduction technique CPU only).
                split_score_func (str): "cosine" or "l2"
                generator_type - (str): candidate generation method "Quantile" or "Uniform".
                feature_weights - (list[float]): Per-feature multiplication weights used when choosing the best split. Weights should be >= 0
            verbose (int, optional): verbosity level. Defaults to 0.
            device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """
        if optimizer is not None:
            if isinstance(optimizer, dict):
                    optimizer = [optimizer]
            optimizer = [setup_optimizer(opt) for opt in optimizer]

        self.optimizer = optimizer
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.verbose = verbose
        self.tree_struct = tree_struct
        self._model = None 
        self.gbrl_params = gbrl_params
        self.device = device
        self.params = None
        if type(self) is GBRL:
            self._model = GBTWrapper(self.input_dim, self.output_dim, self.tree_struct, self.optimizer, self.gbrl_params, self.verbose, self.device)
            self._model.reset()
        self.grad = None
        self.input = None

    def reset_params(self):
        """Resets param attributes
        """
        self.params = None

    def set_bias(self, bias: Union[np.ndarray, th.Tensor]):
        """Sets GBRL bias

        Args:
            bias (Union[np.ndarray, th.Tensor])
        """
        if isinstance(bias, th.Tensor):
            bias = bias.clone().detach().cpu().numpy()
        # GBRL works with 2D numpy arrays.
        if len(bias.shape) == 1:
            bias = bias[:, np.newaxis]
        self._model.set_bias(bias.astype(np.single))

    def set_feature_weights(self, feature_weights: Union[np.ndarray, th.Tensor]):
        """Sets GBRL feature_weights

        Args:
            feature_weights (Union[np.ndarray, th.Tensor])
        """
        if isinstance(feature_weights, th.Tensor):
            feature_weights = feature_weights.clone().detach().cpu().numpy()
        # GBRL works with 2D numpy arrays.
        if len(feature_weights.shape) == 1:
            feature_weights = feature_weights[:, np.newaxis]
        self._model.set_feature_weights(feature_weights.astype(np.single))
    
    def set_bias_from_targets(self, targets: Union[np.ndarray, th.Tensor]):
        """Sets bias as mean of targets

        Args:
            targets (Union[np.ndarray, th.Tensor]): Targets
        """
        if isinstance(targets, th.Tensor):
            arr = targets.clone().detach().cpu().numpy()
        else:
            arr = targets.copy()
        # GBRL works with 2D numpy arrays.
        if len(arr.shape) == 1:
            arr = arr[:, np.newaxis]
        mean_arr = np.mean(arr, axis=0)
        if isinstance(mean_arr, float):
            mean_arr = np.ndarray([mean_arr])
        self._model.set_bias(mean_arr.astype(np.single)) # GBRL only works with float32
    
    def get_iteration(self) -> int:
        """
        Returns:
            int : number of boosting iterations
        """
        return self._model.get_iteration()

    def get_total_iterations(self) -> int:
        """
        Returns:
            int: total number of boosting iterations 
            (sum of actor and critic if they are not shared otherwise equals 
            get_iteration())
        """
        return self._model.get_total_iterations()

    def get_schedule_learning_rates(self) -> Tuple[float, float]:
        """
        Gets learning rate values for optimizers according to schedule of ensemble.
        Constant schedule - no change in values.
        Linear schedule - learning rate value accordign to number of trees in the ensemble.
        Returns:
            Tuple[float, float]: learning rate schedule per optimizer.
        """
        return self._model.get_schedule_learning_rates()

    def step(self,  X: Optional[Union[np.ndarray, th.Tensor]] = None, grad: Optional[Union[np.ndarray, th.Tensor]] = None, max_grad_norm: Optional[float] = None) -> None:
        """Perform a boosting step (fits a single tree on the gradients)

        Args:
            X (Union[np.ndarray, th.Tensor]): inputs
            max_grad_norm (float, optional): perform gradient clipping by norm. Defaults to None.
            grad (Optional[Union[np.ndarray, th.Tensor]], optional): manually calculated gradients. Defaults to None.
        """
        if X is None:
            assert self.input is not None, "Cannot update trees without input. Make sure model is called with requires_grad=True"
            X = self.input
        n_samples = len(X)
        grad = grad if grad is not None else self.params.grad.detach() * n_samples

        grad = clip_grad_norm(grad, max_grad_norm)
        validate_array(grad)
        self._model.step(X, grad)
        self.grad = grad
        self.input = None
        
    def get_params(self) -> Tuple[np.ndarray, np.ndarray]:
        """Returns predicted model parameters and their respective gradients

        Returns:
            Tuple[np.ndarray, np.ndarray]
        """
        assert self.params is not None, "must run a forward pass first"
        params = self.params
        if isinstance(self.params, tuple):
            params = (params[0].detach().cpu().numpy(), params[1].detach().cpu().numpy()) 
        return params, self.grad
    
    def fit(self, X: Union[np.ndarray, th.Tensor], targets: Union[np.ndarray, th.Tensor], iterations: int, shuffle: bool = True, loss_type: str = 'MultiRMSE') -> float:
        """Fit multiple iterations (as in supervised learning)

        Args:
            X (Union[np.ndarray, th.Tensor]): inputs
            targets (Union[np.ndarray, th.Tensor]): targets
            iterations (int): number of boosting iterations
            shuffle (bool, optional): Shuffle dataset. Defaults to True.
            loss_type (str, optional): Loss to use (only MultiRMSE is currently implemented ). Defaults to 'MultiRMSE'.

        Returns:
            float: final loss over all examples.
        """
        return self._model.fit(X, targets, iterations, shuffle, loss_type)
    
    def get_num_trees(self) -> int:
        """
        Returns number of trees in the ensemble

        Returns:
            int: number of trees in the ensemble
        """
        return self._model.get_num_trees()
    
    def tree_shap(self, tree_idx: int, features: Union[np.ndarray, th.Tensor]) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Calculates SHAP values for a single tree
            Implementation based on - https://github.com/yupbank/linear_tree_shap
            See Linear TreeShap, Yu et al, 2023, https://arxiv.org/pdf/2209.08192 

        Args:
            tree_idx (int): tree index
            features (Union[np.ndarray, th.Tensor])

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]: SHAP values of shap [n_samples, number of input features, number of outputs]. The output is a tuple of SHAP values per model only in the case of a separate actor-critic model.
        """
        return self._model.tree_shap(tree_idx, features)
    
    def shap(self, features: Union[np.ndarray, th.Tensor]) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Calculates SHAP values for the entire ensemble
            Implementation based on - https://github.com/yupbank/linear_tree_shap
            See Linear TreeShap, Yu et al, 2023, https://arxiv.org/pdf/2209.08192 
            
        Args:
            features (Union[np.ndarray, th.Tensor])

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]: SHAP values of shap [n_samples, number of input features, number of outputs]. The output is a tuple of SHAP values per model only in the case of a separate actor-critic model.
        """

        return self._model.shap(features)

    def save_model(self, save_path: str) -> None:
        """
        Saves model to file

        Args:
            filename (str): Absolute path and name of save filename.
        """
        self._model.save(save_path) 

    def export_model(self, filename: str, modelname: str = None) -> None:
        """
        Exports model as a C-header file

        Args:
            filename (str): Absolute path and name of exported filename.
        """
        self._model.export(filename, modelname) 

    @classmethod
    def load_model(cls, load_name: str, device: str) -> "GBRL":
        """Loads GBRL model from a file

        Args:
            load_name (str): full path to file name

        Returns:
            GBRL instance 
        """
        instance = cls.__new__(cls)
        instance._model = GBTWrapper.load(load_name, device)
        instance.optimizer =  instance._model.optimizer
        instance.input_dim = instance._model.input_dim
        instance.output_dim = instance._model.output_dim
        instance.verbose = instance._model.verbose
        instance.tree_struct = instance._model.tree_struct
        instance.gbrl_params = instance._model.gbrl_params
        instance.device = instance._model.get_device()
        instance._model.device = instance.device
        return instance

    def set_device(self, device: str):
        """Sets GBRL device (either cpu or cuda)

        Args:
            device (str): choices are ['cpu', 'cuda']
        """
        assert device in ['cpu', 'cuda'], "device must be in ['cpu', 'cuda']"
        self._model.set_device(device)
        self.device = device

    def get_device(self) -> Union[str, Tuple[str, str]]:
        """Returns GBRL device/devices (if multiple GBRL models)

        Returns:
            Union[str, Tuple[str, str]]: GBRL device per model
        """
        return self._model.get_device()

    def __call__(self, X: Union[th.Tensor, np.ndarray], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> Union[th.Tensor, np.ndarray]:
        """Returns GBRL's output as either a Tensor or a numpy array. if `requires_grad=True` then stores 
           differentiable parameters in self.params. 
           Return type/device is identical to the input type/device.

        Args:
            X (Union[th.Tensor, np.ndarray]): Input
            requires_grad (bool, optional). Defaults to True.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Union[th.Tensor, np.ndarray]: Returns model predictions
        """
        y_pred = self._model.predict(X, requires_grad, start_idx, stop_idx, tensor)
        if requires_grad:
            self.grad = None
            self.params = y_pred
            self.input = X
        return y_pred

    def print_tree(self, tree_idx: int) -> None:
        """Prints tree information

        Args:
            tree_idx (int): tree index to print
        """
        self._model.print_tree(tree_idx)

    def plot_tree(self, tree_idx: int, filename: str) -> None:
        """Plots tree using (only works if GBRL was compiled with graphviz)

        Args:
            tree_idx (int): tree index to plot
            filename (str): .png filename to save
        """
        self._model.plot_tree(tree_idx, filename)
    
    def copy(self) -> "GBRL":
        """Copy class instance 

        Returns:
            GradientBoostingTrees: copy of current instance. The actual type will be the type
            of the subclass that calls this method.
        """
        return self.__copy__()
    
    def __copy__(self) -> "GBRL":
        copy_ = GBRL(self.tree_struct.copy(), self.input_dim, self.output_dim, self.optimizer.copy(), self.gbrl_params, self.verbose)
        if self._model is not None:
            copy_._model = self._model.copy()
        return copy_


if __name__ == '__main__':
    pass


##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
# shrinkage is not implemented yet
VALID_OPTIMIZER_ARGS = ['init_lr', 'scheduler', 'shrinkage', 'algo', 'beta_1', 'beta_2', 'eps', 'T', 'start_idx', 'stop_idx']
APPROVED_OPTIMIZERS = ["Adam", "SGD"]



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
import os
from typing import Dict, List, Union, Tuple, Optional, Any

import numpy as np
import torch as th

from gbrl import GBRL_CPP
from gbrl.utils import (get_input_dim, get_poly_vectors, 
                    to_numpy,
                    numerical_dtype, 
                    get_tensor_info,
                    preprocess_features,
                    ensure_same_type,
                    concatenate_arrays,
                    tensor_to_leaf)

class GBTWrapper:
    def __init__(self, input_dim: int, output_dim: int, tree_struct: Dict, optimizer: Union[Dict, List], gbrl_params: Dict, verbose: int = 0, device: str = 'cpu'):
        if 'T' in gbrl_params:
            del gbrl_params['T']
        self.params = {'input_dim': input_dim,
                       'output_dim': output_dim,
                       'split_score_func': gbrl_params.get('split_score_func', 'Cosine'), 
                       'generator_type': gbrl_params.get('generator_type', 'Quantile'), 
                       'use_control_variates': gbrl_params.get('control_variates', False), 
                       'verbose': verbose, 'device': device, **tree_struct}
        self.tree_struct = tree_struct
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.device = device
        self.optimizer = optimizer if isinstance(optimizer, list) or optimizer is None else [optimizer]
        self.student_model = None
        self.cpp_model = None
        self.iteration = 0
        self.total_iterations = 0
        self.gbrl_params = gbrl_params
        self.verbose = verbose
        feature_weights = gbrl_params.get('feature_weights', None)
        if feature_weights is not None:
            feature_weights = to_numpy(feature_weights)
            feature_weights = feature_weights.flatten()
            assert np.all(feature_weights >= 0), "feature weights contains non-positive values"
        else:
            feature_weights = np.ascontiguousarray(np.ones(input_dim, dtype=np.single))
        self.feature_weights = feature_weights


    def reset(self) -> None:
        if self.cpp_model is not None:
            lrs = self.cpp_model.get_scheduler_lrs()
            for i in range(len(self.optimizer)):
                self.optimizer[i]['init_lr'] = lrs[i]

        self.cpp_model = GBRL_CPP(**self.params)
        self.cpp_model.set_feature_weights(self.feature_weights)
        if self.student_model is not None:
            for i in range(len(self.optimizer)):
                self.optimizer[i]['T'] -= self.total_iterations
        else:
            self.total_iterations = 0
        try:
            for opt in self.optimizer:
                self.cpp_model.set_optimizer(**opt)
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")

    def step(self, features: Union[np.ndarray, th.Tensor, Tuple], grads: Union[np.ndarray, th.Tensor]) -> None:
        features, grads = ensure_same_type(features, grads)
        if isinstance(features, th.Tensor):
            if self.feature_weights is None:
                self.feature_weights = th.ones(get_input_dim(features), device=features.device).float()
            self.cpp_model.step(get_tensor_info(features.float()), None, get_tensor_info(grads.float()))
        else:
            num_features, cat_features = preprocess_features(features)
            grads = np.ascontiguousarray(grads.reshape((len(grads), self.params['output_dim']))).astype(numerical_dtype)
            self.cpp_model.step(num_features, cat_features, grads)
        self.iteration = self.cpp_model.get_iteration()
        self.total_iterations += 1

    def fit(self, features: Union[np.ndarray, th.Tensor], targets: Union[np.ndarray, th.Tensor], iterations: int, shuffle: bool = True, loss_type: str = 'MultiRMSE') -> float:
        if isinstance(features, th.Tensor):
            features = features.detach().cpu().numpy()
        num_features, cat_features = preprocess_features(features)
        targets = to_numpy(targets)
        targets = targets.reshape((len(targets), self.params['output_dim'])).astype(numerical_dtype)
        loss = self.cpp_model.fit(num_features, cat_features, targets, iterations, shuffle, loss_type)
        self.iteration = self.cpp_model.get_iteration()
        return loss

    def save(self, filename: str) -> None:
        filename = filename.rstrip('.')
        filename += '.gbrl_model'
        assert self.cpp_model is not None, "Can't save non-existent model!"
        status = self.cpp_model.save(filename)
        assert status == 0, "Failed to save model"

    def export(self, filename: str, modelname: str = None) -> None:
        # exports model to C
        filename = filename.rstrip('.')
        filename += '.h'
        assert self.cpp_model is not None, "Can't export non-existent model!"
        if modelname is None:
            modelname = ""
        try:
            status = self.cpp_model.export(filename, modelname)
            assert status == 0, "Failed to export model"
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")

    @classmethod
    def load(cls, filename: str, device: str) -> "GBTWrapper":
        filename = filename.rstrip('.')
        if '.gbrl_model' not in filename:
            filename += '.gbrl_model'
        assert os.path.isfile(filename), "filename doesn't exist!"
        try:
            instance = cls.__new__(cls)
            instance.cpp_model = GBRL_CPP.load(filename)
            instance.set_device(device)
            metadata =  instance.cpp_model.get_metadata()
            instance.tree_struct = {'max_depth': metadata['max_depth'], 
                            'min_data_in_leaf': metadata['min_data_in_leaf'],
                            'n_bins': metadata['n_bins'], 
                            'par_th': metadata['par_th'],
                            'batch_size': metadata['batch_size'], 
                            'grow_policy': metadata['grow_policy']}
            instance.params = {'input_dim': metadata['input_dim'],
                            'output_dim': metadata['output_dim'],
                            'split_score_func': metadata['split_score_func'], 
                            'generator_type': metadata['generator_type'], 
                            'use_control_variates': metadata['use_control_variates'], 
                            'verbose': metadata['verbose'], 
                            'device': instance.cpp_model.get_device(),
                            **instance.tree_struct
                            }
            instance.output_dim = metadata['output_dim']
            instance.input_dim = metadata['input_dim']
            instance.verbose = metadata['verbose']
            instance.gbrl_params = {'split_score_func': metadata['split_score_func'], 
                                    'generator_type': metadata['generator_type'], 
                                    'use_control_variates': metadata['use_control_variates'], 
                                    }
            instance.optimizer = instance.cpp_model.get_optimizers()
            instance.iteration = metadata['iteration']
            instance.total_iterations = metadata['iteration']
            instance.student_model = None
            instance.feature_weights = instance.cpp_model.get_feature_weights()
            instance.device = instance.params['device']
            return instance
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")
            return None

    def get_schedule_learning_rates(self) -> Union[int, Tuple[int, int]]:
        return self.cpp_model.get_scheduler_lrs()

    def get_total_iterations(self) -> int:
        return self.total_iterations

    def get_iteration(self) -> int:
        return self.cpp_model.get_iteration()
    
    def get_num_trees(self) -> int:
        if self.student_model is not None:
            return self.cpp_model.get_num_trees() + self.student_model.get_num_trees()
        return self.cpp_model.get_num_trees()
    
    def set_bias(self, bias: Union[np.ndarray, float]) -> None:
        if not isinstance(bias, np.ndarray) and not isinstance(bias, float):
            raise TypeError("Input should be a numpy array or float")

        if isinstance(bias, float):
            bias = np.ndarray([float])

        if bias.ndim > 1:
            bias = bias.ravel()
        elif bias.ndim == 0:
            bias = np.ndarray([bias.item()])  # Converts 0D arrays to 1D
        try:
            self.cpp_model.set_bias(bias)
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")

    def set_feature_weights(self, feature_weights: Union[np.ndarray, float]) -> None:
        if not isinstance(feature_weights, np.ndarray) and not isinstance(feature_weights, float):
            raise TypeError("Input should be a numpy array or float")

        if isinstance(feature_weights, float):
            feature_weights = np.array([float])

        if feature_weights.ndim > 1:
            feature_weights = feature_weights.ravel()
        elif feature_weights.ndim == 0:
            feature_weights = np.array([feature_weights.item()])  # Converts 0D arrays to 1D
        assert len(feature_weights) == self.input_dim, "feature weights has to have the same number of elements as features"
        assert np.all(feature_weights >= 0), "feature weights contains non-positive values"
        try:
            self.cpp_model.set_feature_weights(feature_weights)
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")

    def get_bias(self) -> np.ndarray:
        return self.cpp_model.get_bias()
    
    def get_feature_weights(self) -> np.ndarray:
        return self.cpp_model.get_feature_weights()

    def get_device(self) -> str:
        return self.cpp_model.get_device()

    def print_tree(self, tree_idx: int) -> None:
        self.cpp_model.print_tree(tree_idx)
    
    def plot_tree(self, tree_idx: int, filename: str) -> None:
        try:
            self.cpp_model.plot_tree(tree_idx, filename)
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")

    def tree_shap(self, tree_idx: int, features: Union[np.ndarray, th.Tensor]) -> np.ndarray:
        """  
        Implementation based on - https://github.com/yupbank/linear_tree_shap
        See Linear TreeShap, Yu et al, 2023, https://arxiv.org/pdf/2209.08192 
        Args:
            tree_idx (int): tree index
            features (Union[np.ndarray, th.Tensor]):

        Returns:
            np.ndarray: shap values
        """
        if isinstance(features, th.Tensor):
            features = features.detach().cpu().numpy()
        num_features, cat_features = preprocess_features(features)
        base_poly, norm_values, offset = get_poly_vectors(self.params['max_depth'], numerical_dtype)
        return self.cpp_model.tree_shap(tree_idx, num_features, cat_features, np.ascontiguousarray(norm_values), np.ascontiguousarray(base_poly), np.ascontiguousarray(offset)) 
    
    def shap(self, features: Union[np.ndarray, th.Tensor]) -> np.ndarray:
        """  
        Uses Linear tree shap for each tree in the ensemble (sequentially)
        Implementation based on - https://github.com/yupbank/linear_tree_shap
        See Linear TreeShap, Yu et al, 2023, https://arxiv.org/pdf/2209.08192 
        Args:
            features (Union[np.ndarray, th.Tensor]):

        Returns:
            np.ndarray: shap values
        """
        if isinstance(features, th.Tensor):
            features = features.detach().cpu().numpy()
        num_features, cat_features = preprocess_features(features)
        base_poly, norm_values, offset = get_poly_vectors(self.params['max_depth'], numerical_dtype)
        return self.cpp_model.ensemble_shap(num_features, cat_features, np.ascontiguousarray(norm_values), np.ascontiguousarray(base_poly), np.ascontiguousarray(offset)) 
    
    def set_device(self, device: Union[str, th.device]) -> None:
        if isinstance(device, th.device):
            device = device.type
        try:
            self.cpp_model.to_device(device)
            self.device = device
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")
    
    def predict(self, features: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> np.ndarray:
        if stop_idx is None:
            stop_idx = 0
        if isinstance(features, th.Tensor):
            preds_dlpack = self.cpp_model.predict(get_tensor_info(features.float()), None, start_idx, stop_idx)
            preds = th.from_dlpack(preds_dlpack)
            if self.student_model is not None:
                student_dlpack = self.student_model.predict(get_tensor_info(features.float()), None)   
                preds += th.from_dlpack(student_dlpack)
            preds.requires_grad_(requires_grad)
        else:
            num_features, cat_features = preprocess_features(features)
            preds = self.cpp_model.predict(num_features, cat_features, start_idx, stop_idx)
            if self.student_model is not None:
                preds += self.student_model.predict(num_features, cat_features)   
            if tensor: 
                if isinstance(preds, np.ndarray):
                    preds = th.tensor(preds, requires_grad=requires_grad, device=self.device).float()
                else:
                    preds = th.from_dlpack(preds)
                    preds.requires_grad_(requires_grad)
        if not tensor and isinstance(preds, th.Tensor):
            preds = preds.detach().cpu().numpy()
        return preds
    
    def distil(self, obs: Union[np.ndarray, th.Tensor], targets: np.ndarray, params: Dict, verbose: int = 0) -> Tuple[int, Dict]:
        num_obs, cat_obs = preprocess_features(obs)
        distil_params = {'output_dim': self.params['output_dim'], 'split_score_func': 'L2',
                         'generator_type': 'Quantile',  'use_control_variates': False, 'device': self.device,
                        'max_depth': params.get('distil_max_depth', 6), 'verbose': verbose, 'batch_size': self.params.get('distil_batch_size', 2048)}
        self.student_model = GBRL_CPP(**distil_params)
        distil_optimizer = {'algo': 'SGD', 'init_lr': params.get('distil_lr', 0.1)}
        try:
            self.student_model.set_optimizer(**distil_optimizer)
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")

        bias = np.mean(targets, axis=0)
        if isinstance(bias, float):
            bias = np.ndarray([bias])
        self.student_model.set_bias(bias.astype(numerical_dtype))
        tr_loss = self.student_model.fit(num_obs, cat_obs, targets, params['min_steps'])
        while tr_loss> params.get('min_distillation_loss', 0.1):
            if params['min_steps'] < params['limit_steps']:
                steps_to_add = min(500, params['limit_steps'] - params['min_steps'])
                tr_loss = self.student_model.fit(num_obs, cat_obs, targets, steps_to_add, shuffle=False)
                params['min_steps'] += steps_to_add
            else:
                break
        self.reset()
        return tr_loss, params

    def copy(self):
        return self.__copy__()
    
    def print_ensemble_metadata(self):
        self.cpp_model.print_ensemble_metadata()
    
    def __copy__(self):
        copy_ = GBTWrapper(self.input_dim, self.output_dim, self.tree_struct.copy(), [opt.copy() if opt is not None else opt for opt in self.optimizer], self.gbrl_params, self.verbose, self.device)
        copy_.iteration = self.iteration 
        copy_.total_iterations = self.total_iterations
        if self.cpp_model is not None:
            copy_.cpp_model = GBRL_CPP(self.cpp_model)
        if self.student_model is not None:
            copy_.student_model = GBRL_CPP(self.student_model)
        return copy_

class SeparateActorCriticWrapper:
    def __init__(self, input_dim: int, output_dim: int, tree_struct: Dict, policy_optimizer: Dict, value_optimizer: Dict, gbrl_params: Dict, verbose: int = 0, device: str = 'cpu'):
        print('****************************************')
        print(f'Separate GBRL Tree with input dim: {input_dim}, output dim: {output_dim}, tree_struct: {tree_struct} policy_optimizer: {policy_optimizer} value_optimizer: {value_optimizer}')
        print('****************************************')
        self.policy_model = GBTWrapper(input_dim, output_dim - 1, tree_struct, policy_optimizer, gbrl_params, verbose, device)
        self.value_model = GBTWrapper(input_dim, 1, tree_struct, value_optimizer, gbrl_params, verbose, device)
        self.tree_struct = tree_struct
        self.total_iterations = 0
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.policy_optimizer = policy_optimizer
        self.value_optimizer = value_optimizer
        self.gbrl_params = gbrl_params
        self.verbose = verbose
        self.device = device
 
    def step(self, observations: Union[np.ndarray, th.Tensor], theta_grad: Union[np.ndarray, th.Tensor], value_grad: Union[np.ndarray, th.Tensor]):
        self.step_policy(observations, theta_grad)
        self.step_critic(observations, value_grad)
        self.total_iterations += 1
        
    def step_policy(self, observations: Union[np.ndarray, th.Tensor], theta_grad: Union[np.ndarray, th.Tensor]):
        self.policy_model.step(observations, theta_grad)
    
    def step_critic(self, observations: Union[np.ndarray, th.Tensor], value_grad: Union[np.ndarray, th.Tensor]):
        self.value_model.step(observations, value_grad)

    def set_device(self, device:str) -> None:
        self.policy_model.set_device(device)
        self.value_model.set_device(device)
        self.device = device

    def tree_shap(self, tree_idx: int, observations: Union[np.ndarray, th.Tensor]) -> Tuple[np.ndarray, np.ndarray]:
        policy_shap = self.policy_model.tree_shap(tree_idx, observations)
        value_shap = self.value_model.tree_shap(tree_idx, observations)
        return policy_shap, value_shap

    def get_total_iterations(self) -> int:
        return self.total_iterations
    
    def get_device(self) -> Tuple[str, str]:
        return self.policy_model.get_device(), self.value_model.get_device()

    def get_schedule_learning_rates(self) -> Tuple[int, int]:
        policy_lr = self.policy_model.get_schedule_learning_rates()
        value_lr = self.value_model.get_schedule_learning_rates()
        return policy_lr, value_lr
    
    def predict(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        preds = self.policy_model.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        pred_values = self.value_model.predict(observations, requires_grad, start_idx, stop_idx, tensor).squeeze()
        if len(preds.shape) == 1:
            if isinstance(preds, th.Tensor):
                preds = preds.unsqueeze(-1)
            else:
                preds = preds[:, np.newaxis]
        return tensor_to_leaf(preds, requires_grad=requires_grad), tensor_to_leaf(pred_values, requires_grad=requires_grad)

    def predict_policy(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True):
        preds = self.policy_model.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        if len(preds.shape) == 1:
            if isinstance(preds, th.Tensor):
                preds = preds.unsqueeze(-1)
            else:
                preds = preds[:, np.newaxis]
        return preds 
    
    def predict_critic(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True):
        return self.value_model.predict(observations, requires_grad, start_idx, stop_idx, tensor)
    
    def reset(self) -> None:
        self.policy_model.reset()
        self.value_model.reset()

    def save(self, filename: str) -> None:
        self.policy_model.save(filename + '_policy')
        self.value_model.save(filename + '_value')

    def export(self, filename: str) -> None:
        self.policy_model.export(filename + '_policy')
        self.value_model.export(filename + '_value')
    
    def print_tree(self, tree_idx: int) -> None:
        print("Policy ensemble")
        self.policy_model.print_tree(tree_idx)
        print("Value ensemble")
        self.value_model.print_tree(tree_idx)
    
    def plot_tree(self, tree_idx: int, filename: str) -> None:
        print("Policy ensemble")
        self.policy_model.plot_tree(tree_idx, filename.rstrip(".") + "_policy")
        print("Value ensemble")
        self.value_model.plot_tree(tree_idx,  filename.rstrip(".") + "_value")

    @classmethod
    def load(cls, filename: str, device: str) -> "SeparateActorCriticWrapper":
        instance = cls.__new__(cls)
        instance.policy_model = GBTWrapper.load(filename + '_policy', device)
        instance.value_model = GBTWrapper.load(filename + '_value', device)
        instance.policy_model.set_device(device)
        instance.value_model.set_device(device)
        instance.tree_struct = instance.policy_model.tree_struct
        instance.total_iterations = instance.policy_model.iteration + instance.value_model.iteration
        instance.input_dim = instance.policy_model.input_dim
        instance.output_dim = instance.policy_model.output_dim
        instance.policy_optimizer = instance.policy_model.optimizer[0]
        instance.value_optimizer = instance.value_model.optimizer[0]
        instance.gbrl_params = instance.policy_model.gbrl_params
        instance.verbose = instance.policy_model.verbose
        instance.device = instance.policy_model.get_device()
        return instance

    def distil_policy(self, obs: Union[np.ndarray, th.Tensor], targets: np.ndarray, params: Dict) -> Tuple[int, Dict]:
        return self.policy_model.distil(obs, targets, params)

    def distil_value(self, obs: Union[np.ndarray, th.Tensor], targets: np.ndarray, params: Dict) -> Tuple[int, Dict]:
        return self.value_model.distil(obs, targets, params)

    def get_iteration(self) -> Tuple[int, int]:
        return self.policy_model.get_iteration(), self.value_model.get_iteration()
    
    def get_num_trees(self) -> Tuple[int, int]:
        return self.policy_model.get_num_trees(), self.value_model.get_num_trees()
    
    def set_policy_bias(self, bias: np.ndarray) -> None:
        self.policy_model.set_bias(bias)
    
    def set_value_bias(self, bias: np.ndarray) -> None:
        self.value_model.set_bias(bias)
    
    def set_policy_feature_weights(self, feature_weights: np.ndarray) -> None:
        self.policy_model.set_feature_weights(feature_weights)
    
    def set_value_feature_weights(self, feature_weights: np.ndarray) -> None:
        self.value_model.set_feature_weights(feature_weights)

    def get_bias(self) -> Tuple[np.ndarray, np.ndarray]:
        return self.policy_model.get_bias(), self.value_model.get_bias()
    
    def get_feature_weights(self) -> Tuple[np.ndarray, np.ndarray]:
        return self.policy_model.get_feature_weights(), self.value_model.get_feature_weights()

    def copy(self) -> "SeparateActorCriticWrapper":
        return self.__copy__()
    
    def __copy__(self) -> "SeparateActorCriticWrapper":
        copy_ = SeparateActorCriticWrapper(self.input_dim, self.output_dim, self.tree_struct.copy(), self.policy_optimizer.copy(), self.value_optimizer.copy(), self.gbrl_params, self.verbose, self.device)
        copy_.total_iterations = self.total_iterations
        copy_.policy_model = self.policy_model.copy()
        copy_.value_model = self.value_model.copy()
        return copy_
    

class SharedActorCriticWrapper(GBTWrapper):
    def __init__(self, input_dim: int, output_dim: int, tree_struct: Dict, policy_optimizer: Dict, value_optimizer: Dict, gbrl_params: Dict=dict(), verbose: int = 0, device: str = 'cpu'):
        print('****************************************')
        print(f'Shared GBRL Tree with input dim: {input_dim}, output dim: {output_dim}, tree_struct: {tree_struct} policy_optimizer: {policy_optimizer} value_optimizer: {value_optimizer}')
        print('****************************************')
        self.value_optimizer = value_optimizer
        self.policy_optimizer = policy_optimizer
        super().__init__(input_dim, output_dim, tree_struct, policy_optimizer, gbrl_params, verbose, device)

    def reset(self) -> None:
        if self.cpp_model is not None:
            policy_lr, value_lr = self.get_schedule_learning_rates()
            self.policy_optimizer['init_lr'] = policy_lr
            if self.value_optimizer:
                self.value_optimizer['init_lr'] = value_lr
        self.cpp_model = GBRL_CPP(**self.params)
        self.cpp_model.set_feature_weights(self.feature_weights)
        if self.student_model is not None:
            self.policy_optimizer['T'] -= self.total_iterations
        else:
            self.total_iterations = 0
        try:
            self.cpp_model.set_optimizer(**self.policy_optimizer)
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")
        if self.value_optimizer:
            if self.student_model is not None:
                self.value_optimizer['T'] -= self.total_iterations
            try:
                self.cpp_model.set_optimizer(**self.value_optimizer)
            except RuntimeError as e:
                print(f"Caught an exception in GBRL: {e}")
        
    def step(self, observations: Union[np.ndarray, th.Tensor], theta_grad: np.ndarray, value_grad: np.ndarray=None) -> None:
        grads = concatenate_arrays(theta_grad, value_grad)
        observations, grads = ensure_same_type(observations, grads)
        if isinstance(observations, th.Tensor):
            self.cpp_model.step(get_tensor_info(observations.float()), None, get_tensor_info(grads.float()))
        else:
            num_observations, cat_observations = preprocess_features(observations)
            grads = np.ascontiguousarray(grads).astype(numerical_dtype)
            input_dim = 0 if num_observations is None else num_observations.shape[1]
            input_dim += 0 if cat_observations is None else cat_observations.shape[1]
            self.cpp_model.step(num_observations, cat_observations, grads)

        self.iteration = self.cpp_model.get_iteration()
        self.total_iterations += 1

    def distil(self, obs: Union[np.ndarray, th.Tensor], policy_targets: np.ndarray, value_targets: np.ndarray, params: Dict, verbose: int) -> Tuple[float, Dict]:
        targets = policy_targets.squeeze() 
        if self.value_optimizer is not None:
            targets = np.concatenate([policy_targets, value_targets[:, np.newaxis]], axis=1)
        return super().distil(obs, targets, params, verbose)
                                     
    def predict(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        preds = super().predict(observations, requires_grad, start_idx, stop_idx, tensor)
        pred_values = tensor_to_leaf(preds[:, -1], requires_grad=requires_grad)
        preds = tensor_to_leaf(preds[:, :-1], requires_grad=requires_grad)
        return preds, pred_values

    def predict_policy(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True):
        preds, _ = self.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        return preds
    
    def predict_critic(self, observations: Union[np.ndarray, th.Tensor], requires_grad: bool = True, start_idx: int = 0, stop_idx: int = None, tensor: bool = True):
        _, pred_values = self.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        return pred_values
    
    @classmethod
    def load(cls, filename: str, device: str) -> "SharedActorCriticWrapper":
        instance = super().load(filename, device) 
        instance.policy_optimizer = instance.optimizer[0]
        instance.value_optimizer = None if len(instance.optimizer) != 2 else instance.optimizer[1]
        return instance

    def copy(self) -> "SharedActorCriticWrapper":
        return self.__copy__()
    
    def __copy__(self) -> "SharedActorCriticWrapper":
        copy_ = SharedActorCriticWrapper(self.input_dim, self.output_dim, self.tree_struct.copy(), self.policy_optimizer.copy(), None if self.value_optimizer is None else self.value_optimizer.copy(), self.gbrl_params, self.verbose, self.device)
        copy_.iteration = self.iteration 
        copy_.total_iterations = self.total_iterations
        if self.cpp_model is not None:
            copy_.model = GBRL_CPP(self.cpp_model)
        if self.student_model is not None:
            copy_.student_model = GBRL_CPP(self.student_model)
        return copy_


    
        
        




import os
import sys
import subprocess
import platform
import shutil
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils import log
import sysconfig


class CMakeExtension(Extension):
    """Extension to integrate CMake build"""
    def __init__(self, name, sourcedir=''):
        super().__init__(name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)
        print(self.sourcedir)

class CMakeBuild(build_ext):
    """Build extension using CMake"""
    user_options = build_ext.user_options + [
        ('cmake-verbose', None, 'Enable verbose output from CMake'),
    ]

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.cmake_verbose = False

    def finalize_options(self):
        build_ext.finalize_options(self)
        self.cmake_verbose = os.getenv('DEBUG', '0') == '1'

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cfg = 'Debug' if self.cmake_verbose else 'Release'
        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
            '-DPYTHON_EXECUTABLE=' + sys.executable,
            '-DPYTHON_INCLUDE_DIR=' + sysconfig.get_path('include'),
            '-DCMAKE_BUILD_TYPE=' + cfg,
        ]   
        if os.environ.get('COVERAGE', '0') == '1':
             cmake_args.append('-DCOVERAGE=ON')
        if sysconfig.get_config_var('LIBRARY') is not None:
            cmake_args.append('-DPYTHON_LIBRARY=' + sysconfig.get_config_var('LIBRARY'))
        if 'CC' in os.environ:
            cmake_args.append('-DCMAKE_C_COMPILER=' + os.environ['CC'])
        if 'CXX' in os.environ:
            cmake_args.append('-DCMAKE_CXX_COMPILER=' + os.environ['CXX'])
        if 'CPPFLAGS' in os.environ:
            cmake_args.append('-DCMAKE_CXX_FLAGS=' + os.environ['CPPFLAGS'])
        build_args = ['--config', cfg]
        if self.cmake_verbose:
            cmake_args.append('-DCMAKE_VERBOSE_MAKEFILE:BOOL=ON')
            cmake_args.append('--debug-trycompile')
            build_args.append('--verbose')

        if ('CPU_ONLY' not in os.environ and platform.system() != 'Darwin') or ('CPU_ONLY' in os.environ and os.environ['CPU_ONLY'] != '1'):
            cmake_args.append('-DUSE_CUDA=ON')
            if 'CUDACXX' in os.environ:
                cmake_args.append('-DCMAKE_CUDA_COMPILER=' + os.environ['CUDACXX'])
        # # Set CMAKE_PREFIX_PATH for LLVM
        # if platform.system() == 'Darwin':  # MacOS specific logic
        #     brew_prefix = subprocess.check_output(['brew', '--prefix', 'llvm']).decode().strip()
        #     print(brew_prefix)
        #     cmake_args.append(f'-DCMAKE_PREFIX_PATH={brew_prefix}')
        #     cmake_args.append(f'-DLLVM_DIR={brew_prefix}/lib/cmake/llvm')
            
        build_temp = self.build_temp
        if not os.path.exists(build_temp):
            os.makedirs(build_temp)

        # Run cmake configuration
        self.run_subprocess(['cmake', ext.sourcedir] + cmake_args, build_temp)
        # Build the extension
        self.run_subprocess(['cmake', '--build', '.'] + build_args, build_temp)

        self.move_built_library(extdir)

    def run_subprocess(self, cmd, cwd):
        log.info('Running command: {}'.format(' '.join(cmd)))
        try:
            subprocess.check_call(cmd, cwd=cwd)
        except subprocess.CalledProcessError as e:
            log.error(f"Command {cmd} failed with error code {e.returncode}")
            log.error(e.output)
            raise

    def move_built_library(self, build_temp):
        built_objects = []
        for root, _, files in os.walk(build_temp):
            for file in files:
                if file.endswith(('.so', '.pyd', '.dll', '.dylib')):
                    built_objects.append(os.path.join(root, file))

        if not built_objects:
            raise RuntimeError(f"Cannot find built library in {build_temp}")
        for built_object in built_objects:
            dest_path = os.path.join(os.path.dirname(__file__), 'gbrl')
            log.info(f'Moving {built_object} to {dest_path}')
            self.copy_file(built_object, dest_path)

setup(
    name="gbrl",
    version = "1.0.9",
    description = "Gradient Boosted Trees for RL",
    author="Benjamin Fuhrer, Chen Tessler, Gal Dalal",
    author_email="bfuhrer@nvidia.com, ctessler@nvidia.com. gdalal@nvidia.com",
    license="NVIDIA Proprietary Software",
    ext_modules=[CMakeExtension('gbrl/gbrl_cpp', sourcedir='.')],
    cmdclass=dict(build_ext=CMakeBuild),
    packages=find_packages(include=["gbrl"]),  # List of all packages to include
    include_package_data=True,
)





